#include "searchtask.h"

#include "conditioneditor.h"

#include <QWriteLocker>
#include <QReadLocker>
#include <QStandardPaths>
#include <QDateTime>

SearchTask::SearchTask(QObject *parent)
    : QObject{parent}
{
    ukuiSearchTask = new UkuiSearch::UkuiSearchTask;
    ukuiSearchTask->initSearchPlugin(UkuiSearch::SearchProperty::SearchType::File);
    queue = ukuiSearchTask->init();
    ukuiSearchTask->setInformNum(100);
    ukuiSearchTask->setMaxResultNum();

    connect(ukuiSearchTask, &UkuiSearch::UkuiSearchTask::searchError, this, [=](){
        Q_EMIT searchFinished(false);
    });
    connect(ukuiSearchTask, &UkuiSearch::UkuiSearchTask::searchFinished, this, [=](){
        handleSearchResult();
        searchDone = true;
        Q_EMIT searchFinished(true);

        if (searchContentDone || !needSearchFileContent) {
            Q_EMIT finished();
        }
    });

    ukuiSearchContentTask = new UkuiSearch::UkuiSearchTask;
    ukuiSearchTask->initSearchPlugin(UkuiSearch::SearchProperty::SearchType::FileContent);
    contentQueue = ukuiSearchContentTask->init();
    ukuiSearchContentTask->setInformNum(100);
    ukuiSearchContentTask->setMaxResultNum();

    connect(ukuiSearchContentTask, &UkuiSearch::UkuiSearchTask::searchError, this, [=](){
        Q_EMIT searchContentFinished(false);
    });
    connect(ukuiSearchContentTask, &UkuiSearch::UkuiSearchTask::searchFinished, this, [=](){
        handleSearchResult();
        searchContentDone = true;
        Q_EMIT searchContentFinished(true);

        if (searchDone || !needSearchFile) {
            Q_EMIT finished();
        }
    });
}

SearchTask::~SearchTask()
{
    delete queue;
    delete contentQueue;
    ukuiSearchTask->stop();
    ukuiSearchTask->deleteLater();
    ukuiSearchContentTask->stop();
    ukuiSearchContentTask->deleteLater();
}

const QVariant &SearchTask::getConditionsDescription() const
{
    return conditionsDescription;
}

void SearchTask::setConditionsDescription(const QVariant &newConditionsDescription)
{
    if (conditionsDescription == newConditionsDescription)
        return;

    needSearchFile = false;
    needSearchFileContent = false;
    hasContentSearchAndCondition = false;

    conditionsDescription = newConditionsDescription;

    auto newConditions = ConditionEditor::generateContitionsFromDescription(newConditionsDescription);

    rwlock.lockForWrite();
    for (auto oldCondition : conditions) {
        oldCondition->deleteLater();
    }
    conditions = newConditions;

    if (ukuiSearchTask->isSearching(UkuiSearch::SearchProperty::SearchType::File)) {
        ukuiSearchTask->stop();
    }
    ukuiSearchTask->clearAllConditions();

    if (ukuiSearchContentTask->isSearching(UkuiSearch::SearchProperty::SearchType::FileContent)) {
        ukuiSearchContentTask->stop();
    }
    ukuiSearchContentTask->clearAllConditions();

    UkuiSearch::SearchResultProperties properties = {UkuiSearch::SearchProperty::FilePath};
    bool hasDateRangeCondition = false;
    bool hasSearchLocationCondition = false;

    for (auto condition : conditions) {
        switch (condition->type()) {
        case Condition::FileName: {
            if (condition->getFileNameIsContains()) {
                needSearchFile = true;
                auto keywords = condition->getFileNameKeywords();
                for (auto keyword : keywords) {
                    ukuiSearchTask->addKeyword(keyword);
                }
            }
            break;
        }
        case Condition::FileContent: {
            if (condition->isAndCondition()) {
                hasContentSearchAndCondition = true;
            }
            needSearchFileContent = true;
            ukuiSearchContentTask->addKeyword(condition->getFileContentContent());
            break;
        }
        case Condition::SearchLocation: {
            if (condition->getSearchLocationIsContains()) {
                hasDateRangeCondition = true;
                auto path = condition->getSearchLocationPath();
                ukuiSearchTask->addSearchDir(path);
                ukuiSearchContentTask->addSearchDir(path);
            }
            break;
        }
        case Condition::DateRange: {
            // 在线程中查询时间信息
            hasDateRangeCondition = true;
            break;
        }
        default:
            break;
        }
    }

    if (hasContentSearchAndCondition) {
        // 如果包含了与逻辑的文件内容关键字条件，则不进行文件名搜索
        needSearchFile = false;
    }

    // 当条件没有包含文件名和文件内容时，使用文件名搜索进行通配搜索
    if (!needSearchFile && !needSearchFileContent) {
        needSearchFile = true;
        ukuiSearchTask->setMatchAllIfNoKeyword(true);
    }

    rwlock.unlock();

    // 根据条件对搜索task进行设置
    if (hasDateRangeCondition)
        properties.append(UkuiSearch::SearchProperty::ModifiedTime);

    ukuiSearchTask->setResultProperties(UkuiSearch::SearchProperty::SearchType::File, properties);
    ukuiSearchContentTask->setResultProperties(UkuiSearch::SearchProperty::SearchType::FileContent, properties);

    if (!hasSearchLocationCondition) {
        ukuiSearchTask->addSearchDir(QStandardPaths::writableLocation(QStandardPaths::HomeLocation));
        ukuiSearchContentTask->addSearchDir(QStandardPaths::writableLocation(QStandardPaths::HomeLocation));
    }

    ukuiSearchTask->setRecurse(true);
    ukuiSearchContentTask->setRecurse(true);

    // 清空旧条件的结果
    queue->clear();
    contentQueue->clear();
    resultInfos.clear();
    contentResultInfos.clear();

    emit conditionsDescriptionChanged();
}

void SearchTask::start()
{
    if (needSearchFile)
        ukuiSearchTask->startSearch(UkuiSearch::SearchProperty::SearchType::File);

    if (needSearchFileContent)
        ukuiSearchContentTask->startSearch(UkuiSearch::SearchProperty::SearchType::FileContent);
}

void SearchTask::cancel()
{
    if (needSearchFile)
        ukuiSearchTask->stop();

    if (needSearchFileContent)
        ukuiSearchContentTask->stop();
}

void SearchTask::handleSearchResult()
{
    while (!queue->isEmpty()) {
        auto item = queue->dequeue();
        QString path = item.getValue(UkuiSearch::SearchProperty::FilePath).toString();
        QVariant timeVar = item.getValue(UkuiSearch::SearchProperty::ModifiedTime);
        if (timeVar.isValid()) {
            QDateTime modificationTimeDate = timeVar.toDateTime();
            // fixme: 判断日期相关条件
        }
        // fixme: 判断大小相关条件
        // fixme: 判断类型相关条件

        auto info = UKUI::IDM::FileInfo::fromPath(path);
        resultInfos << info;
    }

    while (!contentQueue->isEmpty()) {
        auto item = contentQueue->dequeue();
        QString path = item.getValue(UkuiSearch::SearchProperty::FilePath).toString();
        QVariant timeVar = item.getValue(UkuiSearch::SearchProperty::ModifiedTime);
        if (timeVar.isValid()) {
            QDateTime modificationTimeDate = timeVar.toDateTime();
            // fixme: 判断日期相关条件
        }
        // fixme: 判断大小相关条件
        // fixme: 判断类型相关条件

        auto info = UKUI::IDM::FileInfo::fromPath(path);
        contentResultInfos << info;
    }
}
