#include "fileinfojob.h"
#include "fileinfo_p.h"
#include "condition/condition.h"

#include <peony-qt/global-settings.h>
#include <peony-qt/file-utils.h>
#include <peony-qt/file-meta-info.h>
#include <peony-qt/file-label-model.h>
#include <atomic>

#include <QThreadPool>
#include <QMutexLocker>

#include <QUrl>

#undef signals

#include <gio/gio.h>

namespace UKUI {
namespace IDM {
class FileInfoJobPrivate{
public:
    FileInfoJobPrivate() {
        cancelled = false;
        canFetchInfos = false;
        hasCondition = false;
    }

    QList<QSharedPointer<FileInfo>> infos;
    QList<QSharedPointer<FileInfo>> infosMatchConditions;

    std::atomic_bool cancelled;
    std::atomic_bool canFetchInfos;
    bool hasCondition;
    QList<QSharedPointer<Condition>> andConditions;
    QList<QSharedPointer<Condition>> orConditions;
};
}
}

using namespace UKUI::IDM;

FileInfoJob *FileInfoJob::fromUris(const QStringList &uris, bool addWatchByDefault)
{
    return new FileInfoJob(uris, true, addWatchByDefault);
}

FileInfoJob *FileInfoJob::fromPaths(const QStringList &paths, bool addWatchByDefault)
{
    return new FileInfoJob(paths, false, addWatchByDefault);
}

FileInfoJob *FileInfoJob::fromInfos(const QList<QSharedPointer<FileInfo> > &infos)
{
    auto job = new FileInfoJob;
    job->priv->infos = infos;
    return job;
}

FileInfoJob::~FileInfoJob()
{
    delete priv;
}

void FileInfoJob::start()
{
    // fixme: not use QThreadPool::globalInstance()
    QThreadPool::globalInstance()->start(this);
}

void FileInfoJob::cancel()
{
    priv->cancelled = true;
}

void FileInfoJob::setConditions(const QList<QSharedPointer<Condition> > &conditions)
{
    priv->andConditions.clear();
    priv->orConditions.clear();
    if (conditions.isEmpty()) {
        priv->hasCondition = false;
        return;
    }
    for (auto condition : conditions) {
        if (condition->isAndCondition()) {
            priv->andConditions << condition;
        } else {
            priv->orConditions << condition;
        }
    }
    priv->hasCondition = true;
}

void FileInfoJob::run()
{
    priv->canFetchInfos = false;

    Q_EMIT started();

    for (auto info : priv->infos) {
        GError *error = nullptr;
        g_autoptr (GFile) gfile = g_file_new_for_uri(info->priv->uri.toUtf8().constData());
        g_autoptr (GFileInfo) gfileinfo = g_file_query_info(gfile,
                                                            G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME","
                                                            G_FILE_ATTRIBUTE_STANDARD_ICON","
                                                            G_FILE_ATTRIBUTE_STANDARD_TARGET_URI","
                                                            G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE","
                                                            G_FILE_ATTRIBUTE_STANDARD_SIZE","
                                                            G_FILE_ATTRIBUTE_TIME_ACCESS","
                                                            G_FILE_ATTRIBUTE_TIME_MODIFIED","
                                                            G_FILE_ATTRIBUTE_TIME_CREATED","
                                                            "metadata::peony-file-label-ids",
                                                            G_FILE_QUERY_INFO_NONE, nullptr, &error);
        if (error) {
            QString msg = error->message;
            g_error_free (error);
            {
                QMutexLocker l(&info->priv->mutex);
                info->priv->state = FileInfoPrivate::Invalid;
            }
            Q_EMIT errored(msg);
            continue;
        }
        if (priv->cancelled) {
            Q_EMIT cancelled();
            return;
        }
        g_autofree gchar *path = g_file_get_path(gfile);
        QString displayName = g_file_info_get_attribute_string(gfileinfo, G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME);
        GIcon *gicon = g_file_info_get_icon(gfileinfo);
        QString iconName = Peony::FileUtils::getIconStringFromGIconThreadSafety(gicon);

        guint64 timeAcessMsec = g_file_info_get_attribute_uint64(gfileinfo, G_FILE_ATTRIBUTE_TIME_ACCESS) * 1000;
        guint64 timeModifiedMsec = g_file_info_get_attribute_uint64(gfileinfo, G_FILE_ATTRIBUTE_TIME_MODIFIED) * 1000;
        guint64 timeCreatedMsec = g_file_info_get_attribute_uint64(gfileinfo, G_FILE_ATTRIBUTE_TIME_CREATED) * 1000;

        const gchar *ctype = g_file_info_get_content_type(gfileinfo);
        QString contentType = ctype;
        g_autofree gchar *ftype = g_content_type_get_description(ctype);
        QString fileType = ftype;

        guint64 size = g_file_info_get_size(gfileinfo);
        g_autofree gchar *formatsize = g_format_size(size);

        if (!path) {
            if (g_file_info_has_attribute(gfileinfo, G_FILE_ATTRIBUTE_STANDARD_TARGET_URI)) {
                QUrl url(g_file_info_get_attribute_as_string(gfileinfo, G_FILE_ATTRIBUTE_STANDARD_TARGET_URI));
                path = g_strdup(url.toLocalFile().toUtf8().constData());
            }
        }

        QString dateString;
        if (timeAcessMsec > timeModifiedMsec) {
            // 显示打开时间
            QString dateStringTmp = Peony::GlobalSettings::getInstance()->transToSystemTimeFormat(timeAcessMsec);
            dateString = QObject::tr("Open Date: %1").arg(dateStringTmp);
        } else {
            // fixme: 由于WPS等文件修改时是以临时文件移动方式进行原文件覆盖，所以创建时间和修改时间在保存后是一致的，这里先显示修改时间，后续需要结合版本判断的信息来处理此处显示的内容
            // 显示修改时间
            QString dateStringTmp = Peony::GlobalSettings::getInstance()->transToSystemTimeFormat(timeModifiedMsec);
            dateString = QObject::tr("Modified Date: %1").arg(dateStringTmp);
        }

        QString parentPath = path;
        parentPath.truncate(parentPath.lastIndexOf('/'));

        // 标记信息
        auto metaInfo = Peony::FileMetaInfo::fromGFileInfo(info->priv->uri, gfileinfo);
        auto labelStrings = metaInfo->getMetaInfoStringList("peony-file-label-ids");
        QList<int> labelIds;
        QList<QVariant> labelColors;
        // fixme: 线程安全问题
        auto labelModel = FileLabelModel::getGlobalModel();
        for (QString labelString : labelStrings) {
            int labelId = labelString.toInt();
            auto item = labelModel->itemFromId(labelId);
            if (item) {
                labelColors << item->color();
                labelIds << labelId;
            }
        }

        QString version = "test";
        // thread safety
        {
            QMutexLocker l(&info->priv->mutex);
            info->priv->path = path;
            info->priv->parentPath = parentPath;
            info->priv->name = displayName;
            info->priv->iconName = iconName;
            info->priv->timeAccessMsec = timeAcessMsec;
            info->priv->timeModifiedMsec = timeModifiedMsec;
            info->priv->timeCreatedMsec = timeCreatedMsec;
            info->priv->dateString = dateString;
            info->priv->version = version;
            info->priv->mimeType = contentType;
            info->priv->fileType = fileType;
            info->priv->size = size;
            info->priv->sizeFormat = formatsize;
            info->priv->labelIds = labelIds;
            info->priv->labelColors = labelColors;
            info->priv->state = FileInfoPrivate::State::Valid;
        }

        bool match = true;

        if (priv->hasCondition) {
            // 如果符合所有条件, 则match最终会为true, 如果没有或条件则match初始值为true, 否则为false
            match = priv->orConditions.count() == 0;

            // 或条件, 只需要满足其中之一即可通过
            for (auto condition : priv->orConditions) {
                // 如果存在条件匹配则跳过, 并且设置match为true
                switch (condition->type()) {
                case Condition::FileName: {
                    match = condition->isFileNameMatchedWithKeywords(displayName);
                    break;
                }
//                case Condition::FileContent: {
//                    // 应该在搜索阶段实现文件内容匹配?
//                    break;
//                }
                case Condition::FileSize: {
                    match = condition->isFileSizeMatchedWithSize(size);
                    break;
                }
                case Condition::FileType: {
                    int subType = condition->getFileTypeSelectionType();
                    if (subType < 0) {
                        match = condition->isFileTypeMatchedWithSuffixes(displayName);
                    } else {
                        match = condition->isFileTypeMatchedWithSubType(subType, displayName, contentType);
                    }
                    break;
                }
                case Condition::DateRange: {
                    match = condition->isFileMatchedWithDate(0, timeModifiedMsec);
                    break;
                }
                case Condition::SearchLocation: {
                    match = condition->isFileMatchedWithSearchLocation(path);
                    break;
                }
                default: {
                    match = false;
                    break;
                }
                }

                // 如果有一个条件匹配, 则跳出循环
                if (match) {
                    break;
                }
            }

            if (!match) {
                continue;
            } else {
                // 如果存在不匹配的且条件, 则跳过此次流程
                for (auto condition : priv->andConditions) {
                    // 如果存在条件不匹配则跳过, 并且设置match为false
                    switch (condition->type()) {
                    case Condition::FileName: {
                        match = condition->isFileNameMatchedWithKeywords(displayName);
                        break;
                    }
//                    case Condition::FileContent: {
//                        break;
//                    }
                    case Condition::FileType: {
                        int subType = condition->getFileTypeSelectionType();
                        if (subType < 0) {
                            match = condition->isFileTypeMatchedWithSuffixes(displayName);
                        } else {
                            match = condition->isFileTypeMatchedWithSubType(subType, displayName, contentType);
                        }
                        break;
                    }
                    case Condition::FileSize: {
                        match = condition->isFileSizeMatchedWithSize(size);
                        break;
                    }
                    case Condition::SearchLocation: {
                        match = condition->isFileMatchedWithSearchLocation(path);
                        break;
                    }
                    case Condition::DateRange: {
                        match = condition->isFileMatchedWithDate(0, timeModifiedMsec);
                        break;
                    }
                    default:
                        match = true;
                        break;
                    }

                    // 存在不匹配, 则跳出循环
                    if (!match) {
                        break;
                    }
                }
            }
        }

        if (match) {
            priv->infosMatchConditions << info;
            Q_EMIT infoUpdated(info);
        }
    }

    priv->canFetchInfos = true;

    Q_EMIT finished();
}

QList<QSharedPointer<FileInfo> > FileInfoJob::getInfos()
{
    if (priv->canFetchInfos) {
        if (priv->hasCondition) {
            return priv->infosMatchConditions;
        } else {
            return priv->infos;
        }
    }

    return QList<QSharedPointer<FileInfo>>();
}

FileInfoJob::FileInfoJob(const QStringList &args, bool isUriFormat, bool addWatchByDefault, QObject *parent) : QObject(parent), QRunnable()
{
    setAutoDelete(false);
    priv = new FileInfoJobPrivate;
    if (isUriFormat) {
        for (auto arg : args) {
            priv->infos.append(FileInfo::fromUri(arg, addWatchByDefault));
        }
    } else {
        for (auto arg : args) {
            priv->infos.append(FileInfo::fromPath(arg, addWatchByDefault));
        }
    }
}

FileInfoJob::FileInfoJob()
{
    setAutoDelete(false);
    priv = new FileInfoJobPrivate;
}
