#include "fileinfomanager.h"
#include "fileeventhandlerimpl.h"

#include <QWeakPointer>
#include <QString>
#include <QHash>
#include <QThread>
#include <QEventLoop>
#include <QMutex>
#include <QMutexLocker>
#include <QUrl>
#include <QtRemoteObjects/QRemoteObjectHost>

#include <kysdk/kysdk-system/libkyfilewatcher.hpp>
#include <peony-qt/file-label-model.h>

#include <QDebug>

namespace UKUI {
namespace IDM {

class FileInfoManagerPrivate
{
public:
    FileInfoManagerPrivate();
    ~FileInfoManagerPrivate();

    QHash<QString, QWeakPointer<FileInfo>> hash;
    KYSDK_FILEWATCHER::FileWatcher *watcher;

    QThread *watcherThread;
    QEventLoop *watcherLoop;

    QRemoteObjectHost *host;
    FileEventHandlerImpl *impl;

    QMutex mutex;
    QMutex watcherMutex;
};

FileInfoManagerPrivate::FileInfoManagerPrivate()
{
    watcher = new KYSDK_FILEWATCHER::FileWatcher;

    watcherThread = new QThread;

    watcherLoop = new QEventLoop;

    host = new QRemoteObjectHost;
    impl = new FileEventHandlerImpl;
}

FileInfoManagerPrivate::~FileInfoManagerPrivate()
{
    watcherLoop->quit();
    watcher->clearWatchList();
    watcher->disconnect();
    delete watcher;
    delete watcherLoop;
    delete watcherThread;
}

}
}

using namespace UKUI::IDM;

static FileInfoManager *instance = nullptr;

FileInfoManager *FileInfoManager::globalInstance()
{
    if (!instance) {
        instance = new FileInfoManager;
    }
    return instance;
}

FileInfoManager::FileInfoManager(QObject *parent) : QObject(parent)
{
    // 在主线程中初始化标记model
    FileLabelModel::getGlobalModel();

    priv = new FileInfoManagerPrivate;

    priv->watcher->moveToThread(priv->watcherThread);
    priv->watcherLoop->moveToThread(priv->watcherThread);
    priv->host->moveToThread(priv->watcherThread);
    priv->impl->moveToThread(priv->watcherThread);
    connect(priv->watcherThread, &QThread::started, priv->watcherLoop, [=]{
        priv->host->setHostUrl(QUrl(QString("local:ukui-idm"))); //fixme: add id for user and session
        priv->host->enableRemoting(priv->impl);
        priv->watcherLoop->exec();
    });

    connect(priv->impl, &FileEventHandlerImpl::reciveClientFileEvent, this, &FileInfoManager::remoteFileEvent);

    qRegisterMetaType<FileEventType>("FileEventType");

    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileAccessed, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileAccessed, url, parent);
    });
    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileAttrChanged, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileAttrChanged, url, parent);
    });
    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileClosed, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileClosed, url, parent);
    });
//    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileCreated, this, [=](QString url, QString parent){
//        Q_EMIT fileChanged(FileCreated, url, parent);
//    });
    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileDeleted, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileDeleted, url, parent);
    });
    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileIgnoredAutomatic, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileIgnoredAutomatic, url, parent);
    });
    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileModified, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileModified, url, parent);
    });
    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileMoveIn, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileMoveIn, url, parent);
    });
    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileMoveOut, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileMoveOut, url, parent);
    });
    connect(priv->watcher, &KYSDK_FILEWATCHER::FileWatcher::fileOpened, this, [=](QString url, QString parent){
        Q_EMIT fileChanged(FileOpened, url, parent);
    });

    priv->watcherThread->start();
}

FileInfoManager::~FileInfoManager()
{
    delete priv;
}

QSharedPointer<FileInfo> FileInfoManager::getInfoByUri(const QString &uri)
{
    QMutexLocker l(&priv->mutex);
    auto info = priv->hash.value(uri);
    return info;
}

void FileInfoManager::insertInfo(QSharedPointer<FileInfo> info)
{
    QMutexLocker l(&priv->mutex);
    priv->hash.insert(info->uri(), info);
}

void FileInfoManager::addFileToWatcher(const QString &url, bool *ok)
{
    if (url.isEmpty()) {
        if (ok) {
            *ok = false;
        }
        return;
    }
    QUrl tmp(url);
    QString path = tmp.toLocalFile();
    if (!path.startsWith("/")) {
        if (ok) {
            *ok = false;
            return;
        }
    }
    QMutexLocker l(&priv->watcherMutex);
    int ret = priv->watcher->addWatchTarget(path, KYSDK_FILEWATCHER::FileWatcherType::PERIODIC, KYSDK_FILEWATCHER::FileWatcherAttribute::ALL);
    if (ok) {
        *ok = !ret; // 为0时成功
    }
}

QStringList FileInfoManager::addFilesToWatcher(const QStringList &urls)
{
    QStringList failedFiles;
    for (auto url : urls) {
        bool ok = false;
        addFileToWatcher(url, &ok);
        if (!ok && !url.isEmpty()) {
            failedFiles << url;
        }
    }
    return failedFiles;
}

void FileInfoManager::removeFileFromWatcher(const QString &url, bool *ok)
{
    if (url.isEmpty()) {
        if (ok) {
            *ok = false;
        }
        return;
    }
    QUrl tmp(url);
    QString path = tmp.toLocalFile();
    QMutexLocker l(&priv->watcherMutex);
    int ret = priv->watcher->removeWatchTarget(path);
    if (ok) {
        *ok = !ret; // 为0时成功
    }
}

QStringList FileInfoManager::removeFilesFromWatcher(const QStringList &urls)
{
    QStringList failedFiles;
    for (auto url : urls) {
        bool ok = false;
        removeFileFromWatcher(url, &ok);
        if (!ok && !url.isEmpty()) {
            failedFiles << url;
        }
    }
    return failedFiles;
}
