#include "modelcontroller.h"
#include "spacesfilesfetcher.h"
#include "fileitemmodel.h"
#include "databasemanager.h"
#include "fileitemproxysortfiltermodel.h"

#ifdef COMPLIE_AS_PROJECT
#include "conditioneditor.h"
#include "condition.h"
#include "search/searchtask.h"
#include "fileinfojob.h"
#endif

#include <QDesktopServices>
#include <QMessageBox>

#include <QStandardPaths>
#include <QFile>
#include <QJsonObject>
#include <QJsonDocument>

#include <QMenu>

#include <QThread>

#include <QFileDialog>

#include <QDebug>

static ModelController *instance = nullptr;
static bool modelInited = false;

ModelController *ModelController::globalInstance()
{
    if (!instance) {
        instance = new ModelController;
    }
    return instance;
}

ModelController::ModelController(const QString &dbName, QObject *parent) : QObject(parent)
{
    fetcher = new SpacesFilesFetcher(dbName);
    thread = new QThread;
    fetcher->moveToThread(thread);
    connect(thread, &QThread::started, fetcher, &SpacesFilesFetcher::fetch);
    connect(fetcher, &SpacesFilesFetcher::fetched, this, &ModelController::initModels);
    connect(this, &ModelController::spaceRemoved, this, [=](int spaceId){
        proxyModels.remove(spaceId);
    });

    connect(this, &ModelController::modelCreated, this, [=](FileItemProxySortFilterModel *proxyModel){
        proxyModels.insert(proxyModel->getSpaceId(), proxyModel);
        auto model = qobject_cast<FileItemModel *>(proxyModel->sourceModel());
        connect(this, &ModelController::filesCreated, model, &FileItemModel::handleFilesAdded);
        connect(this, &ModelController::spaceFilesChangedForConditionsChanged, model, &FileItemModel::handleFilesChangedForConditionsChanged);
        connect(this, &ModelController::filesReloaded, model, &FileItemModel::handleFilesReloaded);
        connect(this, &ModelController::spaceRemoved, proxyModel, &FileItemProxySortFilterModel::handleSpaceRemoved);
    });
}

ModelController::~ModelController()
{
    disconnect(thread);
    disconnect(fetcher);
    disconnect(this);
    thread->quit();
    // fixme: 在线程运行中退出会导致崩溃
    thread->wait();
    delete fetcher;
    delete thread;
}

void ModelController::startLoadModel()
{
    if (!thread->isRunning())
        thread->start();
    else {
        qDebug() << "model controller has been started";
        if (modelInited) {
            initModels();
        } else {
            qWarning() << "requets load model while fetcher is loading";
        }
    }
}

void ModelController::startLoadRecentModel()
{
    auto model = new FileItemModel(this);
    auto proxyModel = new FileItemProxySortFilterModel(model, this);
    model->startLoadRecentFiles();
    Q_EMIT recentModelCreated(proxyModel);
}

void ModelController::openFile(const QString &path)
{
    if (path.isEmpty())
        return;
    QUrl url = QUrl::fromLocalFile(path);
    QDesktopServices::openUrl(url);
}

void ModelController::userAddFilesToSpace(const QList<QUrl> &urls, FileItemProxySortFilterModel *model, bool sendSignal)
{
    // fixme: run in thread
    DatabaseManager manager;
    int spaceId = model->getSpaceId();
    for (auto url : urls) {
        QString errMsg;
        QString path = url.toLocalFile();
        bool ok = manager.addFileToSpace(path, spaceId, true, &errMsg);
        if (ok) {
            fetcher->rwLock.lockForRead();
            auto newUserAddedInfos = fetcher->userAddedInfos.value(model->getSpaceId());
            fetcher->rwLock.unlock();
            for (auto url : urls) {
                auto newInfo = UKUI::IDM::FileInfo::fromPath(url.toLocalFile());
                newUserAddedInfos << newInfo;
            }

            fetcher->rwLock.lockForWrite();
            fetcher->userAddedInfos.insert(spaceId, newUserAddedInfos);
            fetcher->rwLock.unlock();
            if (sendSignal) {
                model->handleFileEvent(UKUI::IDM::FileInfoManager::UserAddFile, path, nullptr);
            }
        } else {
            QMessageBox::warning(0, 0, errMsg);
        }
    }
}

void ModelController::removeFileFromSpace(const QString &path, FileItemProxySortFilterModel *model, bool sendSignal)
{
    // fixme: run in thread, and batch process
    DatabaseManager manager;
    QString errMsg1, errMsg2;
    manager.removeFileFromSpace(path, model->getSpaceId(), true, &errMsg1);
    manager.removeFileFromSpace(path, model->getSpaceId(), false, &errMsg2);
    if (errMsg1.isEmpty() || errMsg2.isEmpty() && sendSignal) {
        model->handleFileEvent(UKUI::IDM::FileInfoManager::UserRemoveFile, path, nullptr);
    }
}

void ModelController::removeFileFromAllSpace(const QString &path)
{
    // fixme: run in thread, and batch process
    DatabaseManager manager;
    QString errMsg;
    manager.removeFileFromAllSpace(path, &errMsg);
}

void ModelController::editSpaceCondition(int spaceId)
{
    QString spaceName = fetcher->getSpaceNameById(spaceId);
    qDebug() << "edit space condition" << spaceName;
    Q_EMIT setSpaceConditionRequest(spaceId, spaceName);
}

void ModelController::createSpace()
{
    Q_EMIT createSpaceRequest();
}

void ModelController::removeSpace(int spaceId)
{
    DatabaseManager manager;
    QString errorMsg;
    QString spaceName = fetcher->getSpaceNameById(spaceId);
    if (manager.removeSpace(spaceName, &errorMsg)) {
        fetcher->rwLock.lockForWrite();
        fetcher->spaces.remove(spaceId);
        fetcher->rwLock.unlock();
        qDebug() << "space removed" << spaceId;
        // 同时删除配置文件
        QString jsonPath = QStandardPaths::writableLocation(QStandardPaths::AppConfigLocation) + "/spaces/" + QString::number(spaceId) + ".json";
        QFile file(jsonPath);
        file.remove();
        Q_EMIT spaceRemoved(spaceId);
    } else {
        QMessageBox::warning(0, 0, tr("Failed to remove space %1, error message: %2").arg(spaceName).arg(errorMsg));
    }
}

void ModelController::requestShowMenu(FileItemProxySortFilterModel *model, int index, const QPoint &globalPos)
{
    qDebug() << model->spaceName() << index << globalPos;
    Q_EMIT showMenuRequested(model, index, globalPos);
}

void ModelController::editSearch()
{
    Q_EMIT editSearchRequested();
}

void ModelController::saveSearchResultAsSpace()
{
    Q_EMIT searchSaveRequested();
}

void ModelController::requestAddUserAddedFileToSapce(FileItemProxySortFilterModel *model)
{
    int spaceId = model->getSpaceId();

    if (spaceId >= 0) {
        auto urls = QFileDialog::getOpenFileUrls(nullptr, tr("Choose File to Add"));
        if (!urls.isEmpty()) {
            this->userAddFilesToSpace(urls, model, true);
        }
    } else {
        qWarning() << "request add file to space with invalid spaceId" << spaceId;
    }
}

void ModelController::renameSpace(int spaceId, const QString &spaceName)
{
    DatabaseManager manager;
    QString errorMsg;
    QString oldSpaceName = fetcher->getSpaceNameById(spaceId);
    if (manager.renameSpace(oldSpaceName, spaceName, &errorMsg)) {
        fetcher->rwLock.lockForWrite();
        fetcher->spaces[spaceId] = spaceName;
        fetcher->rwLock.unlock();
        Q_EMIT spaceNameUpdated(spaceId, spaceName);
    } else {
        QMessageBox::warning(0, 0, errorMsg);
    }
}

QMap<int, QString> ModelController::getSpaceIdsWithSpace()
{
    QMap<int, QString> result;
    fetcher->rwLock.lockForRead();
    result = fetcher->spaces;
    fetcher->rwLock.unlock();
    return result;
}

QList<QSharedPointer<UKUI::IDM::FileInfo> > ModelController::getUserAddedFilesFromSpaceById(int spaceId)
{
    return fetcher->getFilesInSpaceById(spaceId, true);
}

FileItemProxySortFilterModel *ModelController::getProxyModelBySpaceId(int spaceId)
{
    return proxyModels.value(spaceId, nullptr);
}

void ModelController::onSpaceConditionChanged(int spaceId, const QString &spaceName, const QVariant conditionsDescription)
{

}

void ModelController::onSpaceCreateRequest(const QString &spaceName, const QVariant &conditionDescription, bool isCreatedFromSearch)
{
    this->onSpaceCreateRequestWithFiles(spaceName, conditionDescription, isCreatedFromSearch, QList<QSharedPointer<UKUI::IDM::FileInfo>>());
}

void ModelController::onSpaceCreateRequestWithFiles(const QString &spaceName, const QVariant &conditionDescription, bool isCreatedFromSearch, QList<QSharedPointer<UKUI::IDM::FileInfo> > infos)
{
    DatabaseManager manager;
    QString errMsg;
    QString realSpaceName;
    bool hasInitFiles = !infos.isEmpty();
    if (manager.createSpace(spaceName, QStringList(), &errMsg, &realSpaceName)) {
        int spaceId = manager.getSpaceIdByName(realSpaceName);
        QJsonObject configObj = conditionDescription.toJsonObject();
        QJsonDocument doc(configObj);
        QString jsonPath = QStandardPaths::writableLocation(QStandardPaths::AppConfigLocation) + "/spaces/" + QString::number(spaceId) + ".json";
        QFile file(jsonPath);
        if (file.open(QFile::WriteOnly|QFile::Text)) {
            file.write(doc.toJson());
            fetcher->rwLock.lockForWrite();
            fetcher->spaces.insert(spaceId, realSpaceName);
            fetcher->rwLock.unlock();
            qDebug() << "space created" << spaceId << realSpaceName;
            spaceCreated(spaceId, realSpaceName);
            auto model = new FileItemModel(spaceId, realSpaceName, infos);
            auto proxyModel = new FileItemProxySortFilterModel(model);
            modelCreated(proxyModel);

            DatabaseManager dbManager;
            for (auto info : infos) {
                dbManager.addFileToSpace(info->path(), realSpaceName);
            }

#ifdef COMPLIE_AS_PROJECT
            if (!isCreatedFromSearch) {
                // 如果此处的调用非搜索保存请求且存在条件，则需要进行搜索和匹配
                auto conditions = ConditionEditor::generateContitionsFromDescription(conditionDescription);
                if (!conditions.isEmpty()) {
                    // 对空间进行按条件进行初始化
                    Q_EMIT this->spaceConditionsChanged(spaceId, realSpaceName, conditionDescription);
                }
            }
#endif
        }
    } else {
        QMessageBox::warning(0, 0, tr("Failed to create space %1, is there a same name space created?").arg(spaceName));
    }
}

void ModelController::initModels()
{
    if (!modelInited) {
        modelInited = true;
    } else {
        for (auto proxyModel : proxyModels) {
            Q_EMIT modelCreated(proxyModel);
        }
        Q_EMIT allModelCreated();
        return;
    }

    auto spaces = fetcher->getAllSpacesName();
    for (auto space : spaces) {
        int id = fetcher->getSpaceIdByName(space);
        auto infos = fetcher->getFilesInSpaceById(id, true);
        int userAddedFilesCount = infos.count();
        auto autoAddedInfos = fetcher->getFilesInSpaceById(id, false);
        infos << autoAddedInfos;
        auto model = new FileItemModel(id, space, infos, this);
        model->autoAddedFilesStartIndex = userAddedFilesCount;
        auto proxyModel = new FileItemProxySortFilterModel(model, this);

        Q_EMIT modelCreated(proxyModel);
    }

    // 从ukui-search获取文件创建事件
    connect(UKUI::IDM::FileInfoManager::globalInstance(), &UKUI::IDM::FileInfoManager::remoteFileEvent, this, [=](int eventType, const QString &filePath){
        if (eventType == 1) {
            Q_EMIT this->fileCreateRequest(filePath);
        }
    });

    // 在线程中处理新增文件和空间条件改变
    connect(this, &ModelController::spaceConditionsChanged, this->fetcher, &SpacesFilesFetcher::updateSpaceConditions);
    connect(this, &ModelController::fileCreateRequest, this->fetcher, &SpacesFilesFetcher::handleFileCreated);
    connect(this, &ModelController::filesReloaded, this->fetcher, &SpacesFilesFetcher::handleAfterFilesReloaded);
    connect(this, &ModelController::spaceRemoved, this->fetcher, &SpacesFilesFetcher::handleSpaceRemoved);

    // 线程结果返回信号
    connect(this->fetcher, &SpacesFilesFetcher::handleFilesCreatedProgressed, this, &ModelController::filesCreated);
    connect(this->fetcher, &SpacesFilesFetcher::handleSpaceConditionsProgressed, this, &ModelController::spaceFilesChangedForConditionsChanged);

    // 处理fetcher的扫描文件请求
    connect(this->fetcher, &SpacesFilesFetcher::requestRescanSpaceWithConditionsDescritpion, this, &ModelController::handleRescanSpaceWithCondition);
    connect(this->fetcher, &SpacesFilesFetcher::requestClearSpace, this, [=](int spaceId){
        Q_EMIT this->filesReloaded(QList<QSharedPointer<UKUI::IDM::FileInfo>>(), spaceId);
    });

    Q_EMIT allModelCreated();
}

void ModelController::handleRescanSpaceWithCondition(int spaceId, const QVariant &conditionsDescription)
{
#ifdef COMPLIE_AS_PROJECT
    // fixme: 操作完成后释放资源
    auto searchTask = new SearchTask;
    searchTask->setConditionsDescription(conditionsDescription);

    connect(searchTask, &SearchTask::finished, this, [=](){
        auto result = searchTask->resultInfos;
        result << searchTask->contentResultInfos;
        auto job = UKUI::IDM::FileInfoJob::fromInfos(result);
        job->setAutoDelete(true);
        job->setConditions(searchTask->conditions);
        connect(job, &UKUI::IDM::FileInfoJob::finished, this, [=](){
            auto realResult = job->getInfos();
            Q_EMIT this->filesReloaded(realResult, spaceId);
        }, Qt::BlockingQueuedConnection);
        job->start();
    });

    connect(this, &ModelController::spaceConditionChangedRequest, searchTask, &SearchTask::cancel);

    searchTask->start();
#endif
}
